# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from json import (
    dumps as dump_json,
    load as load_json,
    loads as parse_json,
)
from os import makedirs
from os.path import (
    exists,
    isdir,
    join,
)
from traceback import format_exc
from typing import (
    Any,
    Callable,
    Dict,
    IO,
    List,
    Optional,
    Set,
)
from xdg import (
    xdg_cache_home,  #  User-specific cache files
    xdg_config_home,  # User-specific configuration files
    xdg_data_home,  #   User-specific data files
)
from deareis.keybindings import Keybinding
from deareis.data.plotting import PlotExportSettings
from deareis.data import (
    DRTSettings,
    FitSettings,
    SimulationSettings,
    TestSettings,
)
from deareis.enums import (
    Action,
    Context,
    action_contexts,
)
import deareis.themes as themes
from deareis.signals import Signal
import deareis.signals as signals
from .defaults import (
    DEFAULT_KEYBINDINGS,
    DEFAULT_MARKERS,
    DEFAULT_COLORS,
    DEFAULT_TEST_SETTINGS,
    DEFAULT_FIT_SETTINGS,
    DEFAULT_SIMULATION_SETTINGS,
    DEFAULT_DRT_SETTINGS,
    DEFAULT_PLOT_EXPORT_SETTINGS,
)


VERSION: int = 4


def _parse_v4(dictionary: dict) -> dict:
    # TODO: Update when VERSION is incremented to 4
    return dictionary


def _parse_v3(dictionary: dict) -> dict:
    dictionary.update(
        {
            "version": 4,
            "default_drt_settings": dictionary.get(
                "default_drt_settings",
                DEFAULT_DRT_SETTINGS.to_dict(),
            ),
            "default_plot_export_settings": dictionary.get(
                "default_plot_export_settings",
                DEFAULT_PLOT_EXPORT_SETTINGS.to_dict(),
            ),
        }
    )
    del dictionary["export_units"]
    del dictionary["export_width"]
    del dictionary["export_height"]
    del dictionary["export_dpi"]
    del dictionary["export_preview"]
    del dictionary["export_title"]
    del dictionary["export_legend"]
    del dictionary["export_legend_location"]
    del dictionary["export_grid"]
    del dictionary["export_tight"]
    del dictionary["export_num_per_decade"]
    del dictionary["export_extension"]
    del dictionary["export_experimental_clear_registry"]
    del dictionary["export_experimental_disable_previews"]
    return _parse_v4(dictionary)


def _parse_v2(dictionary: dict) -> dict:
    return _parse_v3(
        {
            "version": 3,
            "auto_backup_interval": dictionary["auto_backup_interval"],
            "num_per_decade_in_simulated_lines": dictionary[
                "num_per_decade_in_simulated_lines"
            ],
            "markers": dictionary["markers"],
            "colors": dictionary["colors"],
            "default_test_settings": dictionary["default_test_settings"],
            "default_fit_settings": dictionary["default_fit_settings"],
            "default_drt_settings": dictionary.get(
                "default_drt_settings",
                DEFAULT_DRT_SETTINGS.to_dict(),
            ),
            "default_simulation_settings": dictionary["default_simulation_settings"],
            "export_units": dictionary.get("export_units", 1),
            "export_width": dictionary.get("export_width", 10.0),
            "export_height": dictionary.get("export_height", 6.0),
            "export_dpi": dictionary.get("export_dpi", 100),
            "export_preview": dictionary.get("export_preview", 4),
            "export_title": dictionary.get("export_title", True),
            "export_legend": dictionary.get("export_legend", True),
            "export_legend_location": dictionary.get("export_legend_location", 0),
            "export_grid": dictionary.get("export_grid", False),
            "export_tight": dictionary.get("export_tight", False),
            "export_num_per_decade": dictionary.get("export_num_per_decade", 100),
            "export_extension": dictionary.get("export_extension", 4),
            "export_experimental_clear_registry": dictionary.get(
                "export_experimental_clear_registry",
                True,
            ),
            "export_experimental_disable_previews": dictionary.get(
                "export_experimental_disable_previews",
                False,
            ),
        }
    )


def _parse_v1(dictionary: dict) -> dict:
    new_color_marker_keys: Dict[str, str] = {
        "real_error": "residuals_real",
        "imag_error": "residuals_imaginary",
        "nyquist_sim": "nyquist_simulation",
        "bode_magnitude_sim": "bode_magnitude_simulation",
        "bode_phase_sim": "bode_phase_simulation",
        "exploratory_mu_criterion": "mu_Xps_mu_criterion",
        "exploratory_mu": "mu_Xps_mu",
        "exploratory_mu_highlight": "mu_Xps_mu_highlight",
        "exploratory_xps": "mu_Xps_Xps",
        "exploratory_xps_highlight": "mu_Xps_Xps_highlight",
    }
    old: str
    new: str
    colors: dict = dictionary.get("colors", {})
    markers: dict = dictionary.get("markers", {})
    for old, new in new_color_marker_keys.items():
        if old in colors:
            colors[new] = colors[old]
            del colors[old]
        if old in markers:
            markers[new] = markers[old]
            del markers[old]
    return _parse_v2(
        {
            "version": 2,
            "auto_backup_interval": dictionary.get("auto_backup_interval", 10),
            "num_per_decade_in_simulated_lines": dictionary[
                "num_per_decade_in_simulated_lines"
            ],
            "markers": dictionary["markers"],
            "colors": dictionary["colors"],
            "default_test_settings": dictionary["default_test_settings"],
            "default_fit_settings": dictionary["default_fit_settings"],
            "default_drt_settings": DEFAULT_DRT_SETTINGS.to_dict(),
            "default_simulation_settings": dictionary["default_simulation_settings"],
        }
    )


class Config:
    def __init__(self):
        self.config_dir_path: str = join(xdg_config_home(), "DearEIS")
        self.config_path: str = join(self.config_dir_path, "config.json")
        self.auto_backup_interval: int = None  # type: ignore
        self.num_per_decade_in_simulated_lines: int = None  # type: ignore
        self.default_test_settings: TestSettings = None  # type: ignore
        self.default_fit_settings: FitSettings = None  # type: ignore
        self.default_drt_settings: DRTSettings = None  # type: ignore
        self.default_simulation_settings: SimulationSettings = None  # type: ignore
        self.default_plot_export_settings: PlotExportSettings = None  # type: ignore
        self.markers: Dict[str, int] = None  # type: ignore
        self.colors: Dict[str, List[float]] = None  # type: ignore
        self.keybindings: List[Keybinding] = None  # type: ignore
        self.from_dict(self.default_settings())
        if not exists(self.config_path):
            self.save()
        else:
            try:
                self.load()
            except AssertionError:
                signals.emit(
                    Signal.SHOW_ERROR_MESSAGE,
                    traceback=format_exc(),
                    message="Encountered malformed config! Using defaults...",
                )
            except Exception:
                signals.emit(
                    Signal.SHOW_ERROR_MESSAGE,
                    traceback=format_exc(),
                    message="Encountered issue while parsing config! Using defaults...",
                )

    def default_settings(self) -> dict:
        return {
            "version": VERSION,
            "auto_backup_interval": 10,
            "num_per_decade_in_simulated_lines": 100,
            "default_test_settings": DEFAULT_TEST_SETTINGS.to_dict(),
            "default_fit_settings": DEFAULT_FIT_SETTINGS.to_dict(),
            "default_drt_settings": DEFAULT_DRT_SETTINGS.to_dict(),
            "default_plot_export_settings": DEFAULT_PLOT_EXPORT_SETTINGS.to_dict(),
            "default_simulation_settings": DEFAULT_SIMULATION_SETTINGS.to_dict(),
            "colors": DEFAULT_COLORS,
            "markers": DEFAULT_MARKERS,
            "keybindings": list(map(lambda _: _.to_dict(), DEFAULT_KEYBINDINGS)),
        }

    def to_dict(self) -> dict:
        return parse_json(
            dump_json(  # This is done to get new instances in memory
                {
                    "version": VERSION,
                    "auto_backup_interval": self.auto_backup_interval,
                    "num_per_decade_in_simulated_lines": self.num_per_decade_in_simulated_lines,
                    "default_test_settings": self.default_test_settings.to_dict(),
                    "default_fit_settings": self.default_fit_settings.to_dict(),
                    "default_drt_settings": self.default_drt_settings.to_dict(),
                    "default_simulation_settings": self.default_simulation_settings.to_dict(),
                    "default_plot_export_settings": self.default_plot_export_settings.to_dict(),
                    "colors": self.colors,
                    "markers": self.markers,
                    "keybindings": list(map(lambda _: _.to_dict(), self.keybindings)),
                }
            )
        )

    def save(self):
        if not isdir(self.config_dir_path):
            makedirs(self.config_dir_path)
        old_config: str = ""
        fp: IO
        if exists(self.config_path):
            with open(self.config_path, "r") as fp:
                old_config = fp.read().strip()
        new_config: str = dump_json(
            self.to_dict(),
            sort_keys=True,
            indent=2,
        ).strip()
        if new_config == old_config:
            return
        with open(self.config_path, "w") as fp:
            fp.write(new_config)

    def from_dict(self, settings: dict):
        self.auto_backup_interval = settings["auto_backup_interval"]
        self.num_per_decade_in_simulated_lines = settings[
            "num_per_decade_in_simulated_lines"
        ]
        self.default_test_settings = TestSettings.from_dict(
            settings.get(
                "default_test_settings",
                DEFAULT_TEST_SETTINGS.to_dict(),
            )
        )
        self.default_fit_settings = FitSettings.from_dict(
            settings.get(
                "default_fit_settings",
                DEFAULT_FIT_SETTINGS.to_dict(),
            )
        )
        self.default_drt_settings = DRTSettings.from_dict(
            settings.get(
                "default_drt_settings",
                DEFAULT_DRT_SETTINGS.to_dict(),
            )
        )
        self.default_simulation_settings = SimulationSettings.from_dict(
            settings.get(
                "default_simulation_settings",
                DEFAULT_SIMULATION_SETTINGS.to_dict(),
            )
        )
        self.default_plot_export_settings = PlotExportSettings.from_dict(
            settings.get(
                "default_plot_export_settings",
                DEFAULT_PLOT_EXPORT_SETTINGS.to_dict(),
            )
        )
        self.markers = settings["markers"]
        marker_themes: Dict[str, int] = {
            "bode_magnitude_data": themes.bode.magnitude_data,
            "bode_magnitude_simulation": themes.bode.magnitude_simulation,
            "bode_phase_data": themes.bode.phase_data,
            "bode_phase_simulation": themes.bode.phase_simulation,
            "impedance_imaginary_data": themes.impedance.imaginary_data,
            "impedance_imaginary_simulation": themes.impedance.imaginary_simulation,
            "impedance_real_data": themes.impedance.real_data,
            "impedance_real_simulation": themes.impedance.real_simulation,
            "mu_Xps_Xps": themes.mu_Xps.Xps,
            "mu_Xps_mu": themes.mu_Xps.mu,
            "nyquist_data": themes.nyquist.data,
            "nyquist_simulation": themes.nyquist.simulation,
            "residuals_imaginary": themes.residuals.imaginary,
            "residuals_real": themes.residuals.real,
        }
        key: str
        theme: int
        for key, theme in marker_themes.items():
            themes.update_plot_series_theme_marker(theme, self.markers[key])
        self.colors = settings["colors"]
        color_themes: Dict[str, int] = {
            "bode_magnitude_data": themes.bode.magnitude_data,
            "bode_magnitude_simulation": themes.bode.magnitude_simulation,
            "bode_phase_data": themes.bode.phase_data,
            "bode_phase_simulation": themes.bode.phase_simulation,
            "drt_credible_intervals": themes.drt.credible_intervals,
            "drt_imaginary_gamma": themes.drt.imaginary_gamma,
            "drt_mean_gamma": themes.drt.mean_gamma,
            "drt_real_gamma": themes.drt.real_gamma,
            "impedance_imaginary_data": themes.impedance.imaginary_data,
            "impedance_imaginary_simulation": themes.impedance.imaginary_simulation,
            "impedance_real_data": themes.impedance.real_data,
            "impedance_real_simulation": themes.impedance.real_simulation,
            "mu_Xps_Xps": themes.mu_Xps.Xps,
            "mu_Xps_Xps_highlight": themes.mu_Xps.Xps_highlight,
            "mu_Xps_mu": themes.mu_Xps.mu,
            "mu_Xps_mu_criterion": themes.mu_Xps.mu_criterion,
            "mu_Xps_mu_highlight": themes.mu_Xps.mu_highlight,
            "nyquist_data": themes.nyquist.data,
            "nyquist_simulation": themes.nyquist.simulation,
            "residuals_imaginary": themes.residuals.imaginary,
            "residuals_real": themes.residuals.real,
        }
        for key, theme in color_themes.items():
            themes.update_plot_series_theme_color(theme, self.colors[key])
        self.keybindings = list(map(Keybinding.from_dict, settings["keybindings"]))
        self.validate_keybindings(self.keybindings)

    def check_type(self, user: Any, default: Any, key: str):
        assert type(user) == type(default), (user, default, key)
        if type(default) is list:
            user_types: set = set(list(map(type, user)))
            default_types: set = set(list(map(type, default)))
            assert user_types == default_types, (user_types, default_types, key)
        elif type(default) is dict:
            for key in default:
                if key in user:
                    self.check_type(user[key], default[key], key)

    def merge_dicts(self, user: dict, default: dict) -> dict:
        result: dict = {}
        key: str
        for key in default.keys():
            value: Optional[Any] = user.get(key)
            if value is None:
                result[key] = default[key]
                continue
            self.check_type(value, default[key], key)
            if type(value) is dict:
                result[key] = self.merge_dicts(value, default[key])
            else:
                result[key] = value
        return result

    def load(self):
        with open(self.config_path, "r") as fp:
            dictionary: dict = load_json(fp)
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_v1,
            2: _parse_v2,
            3: _parse_v3,
            4: _parse_v4,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        dictionary = self.merge_dicts(parsers[version](dictionary), self.to_dict())
        self.from_dict(dictionary)

    def validate_keybindings(self, keybindings: List[Keybinding]):
        stringified: List[str] = list(map(str, keybindings))
        string: str
        for string in set(stringified):
            count: int = stringified.count(string)
            if count == 1:
                continue
            contexts: List[Set[Context]] = list(
                map(
                    lambda _: set(action_contexts[_.action]),
                    filter(lambda _: str(_) == string, keybindings),
                )
            )
            while contexts:
                a = contexts.pop(0)
                for b in contexts:
                    assert (
                        not a.issuperset(b)
                        and not a.issubset(b)
                        and Context.PROGRAM not in a
                        and Context.PROGRAM not in b
                    ), (
                        "The same keybinding has been applied to multiple actions in the same context or in overlapping contexts:\n- "
                        + "\n- ".join(
                            map(repr, filter(lambda _: str(_) == string, keybindings))
                        )
                    )
        actions: List[Action] = list(map(lambda _: _.action, keybindings))
        action: Action
        for action in set(actions):
            assert actions.count(action) == 1, (
                "Multiple keybindings have been applied to the same action:\n- "
                + "\n- ".join(
                    map(repr, filter(lambda _: _.action == action, keybindings))
                )
            )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from time import sleep
from typing import Tuple
from requests import get
from requests.models import Response
from deareis.signals import Signal
import deareis.signals as signals
from deareis.version import PACKAGE_VERSION


def download_json() -> dict:
    url: str = "https://pypi.org/pypi/DearEIS/json"
    res: Response = get(url)
    res.raise_for_status()  # Check if there were any issues
    return res.json()


def parse_version(string: str) -> Tuple[int, int, int]:
    assert string.count(".") == 2, string
    return tuple(map(int, string.split(".")))  # type: ignore


def is_up_to_date(installed: str, pypi: str) -> bool:
    installed_major: int
    installed_minor: int
    installed_patch: int
    installed_major, installed_minor, installed_patch = parse_version(installed)
    pypi_major: int
    pypi_minor: int
    pypi_patch: int
    pypi_major, pypi_minor, pypi_patch = parse_version(pypi)
    if installed_major > pypi_major:
        return True
    elif installed_major < pypi_major:
        return False
    if installed_minor > pypi_minor:
        return True
    elif installed_minor < pypi_minor:
        return False
    if installed_patch > pypi_patch:
        return True
    elif installed_patch < pypi_patch:
        return False
    return True


def perform_update_check():
    signals.emit(
        Signal.SHOW_BUSY_MESSAGE,
        message="Checking PyPI",
    )
    json: dict = download_json()
    assert (
        "info" in json and "version" in json["info"]
    ), "Failed to extract version number from the response from PyPI!"
    pypi_version: str = json["info"]["version"]
    message: str = f"""
Already up-to-date!

     PyPI: {pypi_version}
Installed: {PACKAGE_VERSION}
    """.strip()
    if not is_up_to_date(PACKAGE_VERSION, pypi_version):
        message = f"""
Update available!

     PyPI: {pypi_version}
Installed: {PACKAGE_VERSION}
        """.strip()
    signals.emit(
        Signal.SHOW_BUSY_MESSAGE,
        message=message,
    )
    sleep(5)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import exists, dirname
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import uuid4
import dearpygui.dearpygui as dpg
from deareis.enums import (
    PlotType,
)
from deareis.data import (
    DataSet,
    DRTResult,
    FitResult,
    PlotSettings,
    Project,
    SimulationResult,
    TestResult,
)
from deareis.gui import ProjectTab
from deareis.gui.file_dialog import FileDialog
from deareis.gui.plotting.copy_appearance import CopyPlotAppearance
from deareis.signals import Signal
from deareis.state import STATE
from deareis.themes import get_random_color_marker
from deareis.utility import calculate_window_position_dimensions
import deareis.signals as signals
import deareis.themes as themes


def new_plot_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    existing_labels: List[str] = list(map(lambda _: _.get_label(), project.get_plots()))
    i: int = 1
    label: str = "Plot"
    while label in existing_labels:
        i += 1
        label = f"Plot ({i})"
    settings: PlotSettings = PlotSettings(
        label,
        PlotType.NYQUIST,
        [],
        {},
        {},
        {},
        {},
        {},
        uuid4().hex,
    )
    project.add_plot(settings)
    project_tab.populate_plots(project)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def rename_plot_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    label: Optional[str] = kwargs.get("label")
    if settings is None or label is None:
        return
    project.edit_plot_label(settings, label)
    project_tab.populate_plots(project)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def rename_plot_series(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    label: Optional[str] = kwargs.get("label")
    settings: Optional[PlotSettings] = kwargs.get("settings")
    uuid: Optional[str] = kwargs.get("uuid")
    series: Optional[Union[DataSet, TestResult, FitResult, SimulationResult]]
    series = kwargs.get("series")
    if label is None or settings is None or uuid is None or series is None:
        return
    settings.set_series_label(uuid, label)
    if label == "":
        label = series.get_label()
    elif label.strip() == "":
        label = ""
    project_tab.set_series_label(uuid, label)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def reorder_plot_series(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    uuid: Optional[str] = kwargs.get("uuid")
    step: Optional[int] = kwargs.get("step")
    if step is None or settings is None or uuid is None:
        return
    assert abs(step) == 1
    assert uuid in settings.series_order
    index: int = settings.series_order.index(uuid)
    if index + step < 0 or index + step > len(settings.series_order) - 1:
        return
    settings.series_order.insert(index + step, settings.series_order.pop(index))
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings, adjust_limits=False)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


# TODO: Refactor into a class(?)
def modify_plot_series_theme(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    series: Optional[
        Union[DataSet, TestResult, FitResult, SimulationResult]
    ] = kwargs.get("series")
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if series is None or settings is None:
        return
    window: int = dpg.generate_uuid()
    handler: int = dpg.generate_uuid()
    states: List[str] = []

    def accept_plot_series_theme():
        assert type(states) is list and len(states) == 2
        dpg.hide_item(window)
        dpg.delete_item(window)
        dpg.delete_item(handler)
        before: str
        after: str
        before, after = states
        if before == after:
            return
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)

    uuid: str = series.uuid
    marker_lookup: Dict[int, str] = {v: k for k, v in themes.PLOT_MARKERS.items()}
    color: List[float] = settings.get_series_color(uuid)
    marker: int = settings.get_series_marker(uuid)
    show_line: bool = settings.get_series_line(uuid)

    def hash_state() -> str:
        return f"{','.join(map(str, color))};{str(marker)};{str(show_line)}"

    states.extend([hash_state(), hash_state()])

    def update_color(new_color: List[float]):
        nonlocal states
        nonlocal color
        new_color = list(map(lambda _: _ * 255.0, new_color))
        color = new_color[:]
        settings.set_series_color(uuid, new_color)  # type: ignore
        states[1] = hash_state()

    color_edit: int = dpg.generate_uuid()

    def randomize_color():
        nonlocal states
        nonlocal color
        new_color: List[float]
        new_color, _ = get_random_color_marker({})
        while new_color == color:
            new_color, _ = get_random_color_marker({})
        color = new_color[:]
        settings.set_series_color(uuid, new_color)  # type: ignore
        states[1] = hash_state()
        dpg.set_value(color_edit, color)

    def update_marker(label: str):
        assert project is not None
        assert project_tab is not None
        assert settings is not None
        nonlocal states
        nonlocal marker
        marker = themes.PLOT_MARKERS.get(label, -1)
        settings.set_series_marker(uuid, marker)  # type: ignore
        project_tab.update_plots(
            settings,
            project.get_data_sets(),
            project.get_all_tests(),
            project.get_all_drts(),
            project.get_all_fits(),
            project.get_simulations(),
        )
        states[1] = hash_state()

    def update_line(state: bool):
        assert project is not None
        assert project_tab is not None
        assert settings is not None
        nonlocal states
        nonlocal show_line
        show_line = state
        settings.set_series_line(uuid, state)  # type: ignore
        project_tab.update_plots(
            settings,
            project.get_data_sets(),
            project.get_all_tests(),
            project.get_all_drts(),
            project.get_all_fits(),
            project.get_simulations(),
        )
        states[1] = hash_state()

    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(200, 100)
    x, y = dpg.get_mouse_pos()
    y += 80
    with dpg.window(
        label="Edit appearance",
        modal=True,
        no_move=True,
        no_resize=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        on_close=accept_plot_series_theme,
        tag=window,
    ):
        with dpg.group(horizontal=True):
            dpg.add_text(" Color")
            dpg.add_color_edit(
                default_value=color,
                alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                no_inputs=True,
                alpha_bar=True,
                callback=lambda s, a, u: update_color(a),
                tag=color_edit,
            )
            dpg.add_button(
                label="Randomize",
                callback=randomize_color,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Marker")
            dpg.add_combo(
                items=["None"] + list(themes.PLOT_MARKERS.keys()),
                default_value=marker_lookup.get(marker, "None"),
                width=-1,
                callback=lambda s, a, u: update_marker(a),
            )
        with dpg.group(horizontal=True):
            dpg.add_text("  Line")
            dpg.add_checkbox(
                default_value=settings.get_series_line(uuid),
                callback=lambda s, a, u: update_line(a),
            )
    dpg.bind_item_theme(window, themes.transparent_modal_background)

    with dpg.handler_registry(tag=handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=accept_plot_series_theme,
        )
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)


def export_plot(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None:
        return
    # TODO: Get current x- and y-limits
    STATE.show_plot_exporter(settings, project)


def save_plot(*args, **kwargs):
    fig = kwargs["figure"]  # Optional[matplotlib.figure.Figure]
    if fig is None:
        return
    STATE.close_plot_exporter()

    def save(*args, **kwargs):
        path: str = kwargs["path"]
        directory: str = dirname(path)
        if exists(directory):
            STATE.latest_plot_directory = directory
        fig.savefig(path)

    FileDialog(
        cwd=STATE.latest_plot_directory,
        label="Select file path",
        callback=lambda *a, **k: save(*a, **k),
        default_extension=kwargs["default_extension"],
        extensions=kwargs["extensions"],
        save=True,
    )


def select_plot_appearance_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None or len(settings.series_order) == 0:
        return
    if len(project.get_plots()) < 2:
        return
    CopyPlotAppearance(settings, project)


def copy_plot_appearance_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    changes: Optional[Dict[str, Tuple[str, List[float], int, bool]]] = kwargs.get(
        "changes"
    )
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if changes is None or settings is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Updating plots")
    uuid: str
    label: str
    color: List[float]
    marker: str
    line: bool
    for uuid, (label, color, marker, line) in changes.items():
        settings.set_series_label(uuid, label)
        settings.set_series_color(uuid, color)
        themes.update_plot_series_theme_color(
            settings.get_series_theme(uuid), settings.get_series_color(uuid)
        )
        settings.set_series_marker(uuid, marker)
        themes.update_plot_series_theme_marker(
            settings.get_series_theme(uuid), settings.get_series_marker(uuid)
        )
        settings.set_series_line(uuid, line)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def select_plot_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None:
        return
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Updating plots")
    project_tab.select_plot(
        settings,
        project.get_data_sets(),
        project.get_all_tests(),
        project.get_all_drts(),
        project.get_all_fits(),
        project.get_simulations(),
        adjust_limits=kwargs.get("adjust_limits", True),
        plot_only=kwargs.get("plot_only", False),
    )
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def select_plot_type(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None:
        return
    plot_type: Optional[PlotType] = kwargs.get("plot_type")
    if plot_type is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Updating plots")
    settings.set_type(plot_type)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings, plot_only=True)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def toggle_plot_series(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Updating plots")
    enabled: bool = kwargs.get("enabled", False)
    data_sets: Optional[List[DataSet]] = kwargs.get("data_sets")
    tests: Optional[List[TestResult]] = kwargs.get("tests")
    drts: Optional[List[DRTResult]] = kwargs.get("drts")
    fits: Optional[List[FitResult]] = kwargs.get("fits")
    simulations: Optional[List[SimulationResult]] = kwargs.get("simulations")
    if data_sets is not None:
        if enabled:
            list(map(settings.add_series, data_sets))
        else:
            list(map(lambda _: settings.remove_series(_.uuid), data_sets))
    if tests is not None:
        if enabled:
            list(map(settings.add_series, tests))
        else:
            list(map(lambda _: settings.remove_series(_.uuid), tests))
    if drts is not None:
        if enabled:
            list(map(settings.add_series, drts))
        else:
            list(map(lambda _: settings.remove_series(_.uuid), drts))
    if fits is not None:
        if enabled:
            list(map(settings.add_series, fits))
        else:
            list(map(lambda _: settings.remove_series(_.uuid), fits))
    if simulations is not None:
        if enabled:
            list(map(settings.add_series, simulations))
        else:
            list(map(lambda _: settings.remove_series(_.uuid), simulations))
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def delete_plot_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[PlotSettings] = kwargs.get("settings")
    if settings is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting plot")
    project.delete_plot(settings)
    plots: List[PlotSettings] = project.get_plots()
    if not plots:
        signals.emit(Signal.NEW_PLOT_SETTINGS)
    else:
        project_tab.populate_plots(project)
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=plots[0])
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Tuple,
    Optional,
)
from numpy import (
    angle,
    array,
    floating,
    issubdtype,
    ndarray,
)
from scipy.signal import find_peaks
from pandas import DataFrame
from pyimpspec import (
    Circuit,
    parse_cdc,
)
from deareis.enums import (
    DRTMethod,
    DRTMode,
    RBFShape,
    RBFType,
    drt_method_to_label,
    drt_mode_to_label,
    label_to_drt_method,
    label_to_drt_mode,
    label_to_rbf_shape,
    label_to_rbf_type,
    rbf_shape_to_label,
    rbf_type_to_label,
)
from deareis.utility import format_timestamp


VERSION: int = 2


def _parse_settings_v2(dictionary: dict) -> dict:
    # TODO: Update implementation once VERSION is incremented
    return dictionary


def _parse_settings_v1(dictionary: dict) -> dict:
    dictionary["circuit"] = ""
    dictionary["W"] = 0.15
    dictionary["num_per_decade"] = 100
    return _parse_settings_v2(dictionary)


@dataclass(frozen=True)
class DRTSettings:
    """
    The settings to use when performing a DRT analysis.

    Parameters
    ----------
    method: DRTMethod
        The method to use to perform the analysis.

    mode: DRTMode
        The mode or type of data  (i.e., complex, real, or imaginary) to use.
        TR-NNLS and TR-RBF methods only.

    lambda_value: float
        The Tikhonov regularization parameter to use.
        TR-NNLS and TR-RBF methods only.

    rbf_type: RBFType
        The radial basis function to use for discretization.
        BHT and TR-RBF methods only.

    derivative_order: int
        The derivative order to use when calculating the penalty in the Tikhonov regularization.
        BHT and TR-RBF methods only.

    rbf_shape: RBFShape
        The shape to use with the radial basis function discretization.
        BHT and TR-RBF methods only.

    shape_coeff: float
        The shape coefficient.
        BHT and TR-RBF methods only.

    inductance: bool
        Whether or not to include an inductive term in the calculations.
        TR-RBF methods only.

    credible_intervals: bool
        Whether or not to calculate Bayesian credible intervals.
        TR-RBF methods only.

    num_samples: int
        The number of samples to use when calculating:
        - the Bayesian credible intervals (TR-RBF method)
        - the Jensen-Shannon distance (BHT method)

    num_attempts: int
        The number of attempts to make to find a solution.
        BHT method only.

    maximum_symmetry: float
        The maximum vertical peak-to-peak symmetry allowed.
        Used to discard results with strong oscillations.
        Smaller values provide stricter conditions.
        BHT and TR-RBF methods only.

    circuit: Optional[Circuit]
        A circuit that contains one or more "(RQ)" or "(RC)" elements connected in series.
        An optional series resistance may also be included.
        For example, a circuit with a CDC representation of "R(RQ)(RQ)(RC)" would be a valid circuit.
        It is highly recommended that the provided circuit has already been fitted.
        However, if all of the various parameters of the provided circuit are at their default values, then an attempt will be made to fit the circuit to the data.
        m(RQ)fit method only.

    W: float
        The width of the Gaussian curve that is used to approximate the DRT of an "(RC)" element.
        m(RQ)fit method only.

    num_per_decade: int
        The number of points per decade to use when calculating a DRT.
        m(RQ)fit method only.
    """

    method: DRTMethod
    mode: DRTMode
    lambda_value: float
    rbf_type: RBFType
    derivative_order: int
    rbf_shape: RBFShape
    shape_coeff: float
    inductance: bool
    credible_intervals: bool
    num_samples: int
    num_attempts: int
    maximum_symmetry: float
    circuit: Optional[Circuit]
    W: float
    num_per_decade: int

    def __repr__(self) -> str:
        return f"DRTSettings ({hex(id(self))})"

    def to_dict(self) -> dict:
        return {
            "version": VERSION,
            "method": int(self.method),
            "mode": int(self.mode),
            "lambda_value": self.lambda_value,
            "rbf_type": int(self.rbf_type),
            "derivative_order": self.derivative_order,
            "rbf_shape": int(self.rbf_shape),
            "shape_coeff": self.shape_coeff,
            "inductance": self.inductance,
            "credible_intervals": self.credible_intervals,
            "num_samples": self.num_samples,
            "num_attempts": self.num_attempts,
            "maximum_symmetry": self.maximum_symmetry,
            "circuit": self.circuit.to_string(12) if self.circuit is not None else "",
            "W": self.W,
            "num_per_decade": self.num_per_decade,
        }

    @classmethod
    def from_dict(Class, dictionary: dict) -> "DRTSettings":
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_settings_v1,
            2: _parse_settings_v2,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        del dictionary["version"]
        dictionary = parsers[version](dictionary)
        dictionary["method"] = DRTMethod(dictionary["method"])
        dictionary["mode"] = DRTMode(dictionary["mode"])
        dictionary["rbf_type"] = RBFType(dictionary["rbf_type"])
        dictionary["rbf_shape"] = RBFShape(dictionary["rbf_shape"])
        dictionary["circuit"] = (
            parse_cdc(dictionary["circuit"]) if dictionary["circuit"] != "" else None
        )
        return Class(**dictionary)


def _parse_result_v2(dictionary: dict) -> dict:
    # TODO: Update implementation once VERSION is incremented
    return dictionary


def _parse_result_v1(dictionary: dict) -> dict:
    return _parse_result_v2(dictionary)


@dataclass
class DRTResult:
    """
    An object representing the results of calculating the distribution of relaxation times in a  data set.

    Parameters
    ----------
    uuid: str
        The universally unique identifier assigned to this result.

    timestamp: float
        The Unix time (in seconds) for when the test was performed.

    tau: ndarray
        The time constants (in seconds).

    gamma: ndarray
        The corresponding gamma(tau) values (in ohms).
        These are the gamma(tau) for the real part when the BHT method has been used.

    frequency: ndarray
        The frequencies of the analyzed data set.

    impedance: ndarray
        The modeled impedances.

    real_residual: ndarray
        The residuals for the real parts of the modeled and experimental impedances.

    imaginary_residual: ndarray
        The residuals for the imaginary parts of the modeled and experimental impedances.

    mean_gamma: ndarray
        The mean values for gamma(tau).
        Only non-empty when the TR-RBF method has been used and the Bayesian credible intervals have been calculated.

    lower_bound: ndarray
        The lower bound for the gamma(tau) values.
        Only non-empty when the TR-RBF method has been used and the Bayesian credible intervals have been calculated.

    upper_bound: ndarray
        The upper bound for the gamma(tau) values.
        Only non-empty when the TR-RBF method has been used and the Bayesian credible intervals have been calculated.

    imaginary_gamma: ndarray
        These are the gamma(tau) for the imaginary part when the BHT method has been used.
        Only non-empty when the BHT method has been used.

    scores: Dict[str, complex]
        The scores calculated for the analyzed data set.
        Only non-empty when the BHT method has been used.

    chisqr: float
        The chi-square goodness of fit value for the modeled impedance.

    lambda_value: float
        The regularization parameter used as part of the Tikhonov regularization.
        Only valid (i.e., positive) when the TR-NNLS or TR-RBF methods have been used.

    mask: Dict[int, bool]
        The mask that was applied to the analyzed data set.

    settings: DRTSettings
        The settings used to perform this analysis.
    """

    uuid: str
    timestamp: float
    tau: ndarray
    gamma: ndarray
    frequency: ndarray
    impedance: ndarray
    real_residual: ndarray
    imaginary_residual: ndarray
    mean_gamma: ndarray
    lower_bound: ndarray
    upper_bound: ndarray
    imaginary_gamma: ndarray
    scores: Dict[str, complex]
    chisqr: float
    lambda_value: float
    mask: Dict[int, bool]
    settings: DRTSettings

    def __repr__(self) -> str:
        return f"DRTResult ({self.get_label()}, {hex(id(self))})"

    @classmethod
    def from_dict(Class, dictionary: dict) -> "DRTResult":
        """
        Create an instance from a dictionary.
        """
        assert type(dictionary) is dict
        assert "version" in dictionary
        version: int = dictionary["version"]
        assert version <= VERSION, f"{version=} > {VERSION=}"
        parsers: Dict[int, Callable] = {
            1: _parse_result_v1,
            2: _parse_result_v2,
        }
        assert version in parsers, f"{version=} not in {parsers.keys()=}"
        dictionary = parsers[version](dictionary)
        dictionary["tau"] = array(dictionary["tau"])
        dictionary["gamma"] = array(dictionary["gamma"])
        dictionary["frequency"] = array(dictionary["frequency"])
        dictionary["real_residual"] = array(dictionary["real_residual"])
        dictionary["imaginary_residual"] = array(dictionary["imaginary_residual"])
        dictionary["mean_gamma"] = array(dictionary["mean_gamma"])
        dictionary["lower_bound"] = array(dictionary["lower_bound"])
        dictionary["upper_bound"] = array(dictionary["upper_bound"])
        dictionary["imaginary_gamma"] = array(dictionary["imaginary_gamma"])
        dictionary["settings"] = DRTSettings.from_dict(dictionary["settings"])
        del dictionary["version"]
        mask: Dict[str, bool] = dictionary["mask"]
        key: str
        for key in list(mask.keys()):
            flag: bool = mask[key]
            del mask[key]
            mask[int(key)] = flag
        dictionary["impedance"] = array(
            list(
                map(
                    lambda _: complex(*_),
                    zip(
                        dictionary["real_impedance"],
                        dictionary["imaginary_impedance"],
                    ),
                )
            )
        )
        del dictionary["real_impedance"]
        del dictionary["imaginary_impedance"]
        dictionary["scores"] = {
            k: complex(
                dictionary["real_scores"][k],
                dictionary["imaginary_scores"][k],
            )
            for k in dictionary["real_scores"]
        }
        del dictionary["real_scores"]
        del dictionary["imaginary_scores"]
        return Class(**dictionary)

    def to_dict(self) -> dict:
        """
        Return a dictionary that can be used to recreate an instance.
        """
        dictionary: dict = {
            "version": VERSION,
            "uuid": self.uuid,
            "timestamp": self.timestamp,
            "tau": list(self.tau),
            "gamma": list(self.gamma),
            "real_impedance": list(self.impedance.real),
            "imaginary_impedance": list(self.impedance.imag),
            "frequency": list(self.frequency),
            "real_residual": list(self.real_residual),
            "imaginary_residual": list(self.imaginary_residual),
            "mean_gamma": list(self.mean_gamma),
            "lower_bound": list(self.lower_bound),
            "upper_bound": list(self.upper_bound),
            "imaginary_gamma": list(self.imaginary_gamma),
            "real_scores": {k: v.real for k, v in self.scores.items()},
            "imaginary_scores": {k: v.imag for k, v in self.scores.items()},
            "chisqr": self.chisqr,
            "lambda_value": self.lambda_value,
            "mask": {k: True for k, v in self.mask.items() if v is True},
            "settings": self.settings.to_dict(),
        }
        return dictionary

    def get_label(self) -> str:
        """
        Generate a label for the result.
        """
        method: str = drt_method_to_label[self.settings.method]
        timestamp: str = format_timestamp(self.timestamp)
        return f"{method} ({timestamp})"

    def get_frequency(self) -> ndarray:
        """
        Get the frequencies (in hertz) of the data set.

        Returns
        -------
        ndarray
        """
        return self.frequency

    def get_impedance(self) -> ndarray:
        """
        Get the complex impedance of the model.

        Returns
        -------
        ndarray
        """
        return self.impedance

    def get_tau(self) -> ndarray:
        """
        Get the time constants.

        Returns
        -------
        ndarray
        """
        return self.tau

    def get_gamma(self, imaginary: bool = False) -> ndarray:
        """
        Get the gamma values.

        Parameters
        ----------
        imaginary: bool = False
            Get the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        ndarray
        """
        if imaginary is True:
            return self.imaginary_gamma
        return self.gamma

    def to_dataframe(
        self,
        threshold: float = 0.0,
        imaginary: bool = False,
        latex_labels: bool = False,
        include_frequency: bool = False,
    ) -> DataFrame:
        """
        Get the peaks as a pandas.DataFrame object that can be used to generate, e.g., a Markdown table.

        Parameters
        ----------
        threshold: float = 0.0
            The threshold for the peaks (0.0 to 1.0 relative to the highest peak).

        imaginary: bool = False
            Use the imaginary gamma (non-empty only when using the BHT method).

        latex_labels: bool = False
            Whether or not to use LaTeX macros in the labels.

        include_frequency: bool = False
            Whether or not to also include a column with the frequencies corresponding to the time constants.

        Returns
        -------
        DataFrame
        """
        tau: ndarray
        gamma: ndarray
        tau, gamma = self.get_peaks(threshold=threshold, imaginary=imaginary)
        f: ndarray = 1 / tau
        dictionary: dict = {}
        dictionary["tau (s)" if not latex_labels else r"$\tau$ (s)"] = tau
        if include_frequency is True:
            dictionary["f (Hz)" if not latex_labels else r"$f$ (Hz)"] = f
        dictionary[
            "gamma (ohms)" if not latex_labels else r"$\gamma\ (\Omega)$"
        ] = gamma
        return DataFrame.from_dict(dictionary)

    def get_peaks(
        self,
        threshold: float = 0.0,
        imaginary: bool = False,
    ) -> Tuple[ndarray, ndarray]:
        """
        Get the time constants (in seconds) and gamma (in ohms) of peaks with magnitudes greater than the threshold.
        The threshold and the magnitudes are all relative to the magnitude of the highest peak.

        Parameters
        ----------
        threshold: float = 0.0
            The threshold for the relative magnitude (0.0 to 1.0).

        imaginary: bool = False
            Use the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        assert (
            issubdtype(type(threshold), floating) and 0.0 <= threshold <= 1.0
        ), threshold
        assert type(imaginary) is bool, imaginary
        gamma: ndarray = self.gamma if not imaginary else self.imaginary_gamma
        assert type(gamma) is ndarray, gamma
        if not gamma.any():
            return (
                array([]),
                array([]),
            )
        indices: ndarray
        indices, _ = find_peaks(gamma)
        if not indices.any():
            return (
                array([]),
                array([]),
            )
        max_g: float = max(gamma)
        if max_g == 0.0:
            return (
                array([]),
                array([]),
            )
        indices = array(
            list(
                filter(
                    lambda _: gamma[_] / max_g > threshold and gamma[_] > 0.0, indices
                )
            )
        )
        if indices.any():
            return (
                self.tau[indices],
                gamma[indices],
            )
        return (
            array([]),
            array([]),
        )

    def get_nyquist_data(self) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Nyquist plot: the real and the negative imaginary parts of the impedances.

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        return (
            self.impedance.real,
            -self.impedance.imag,
        )

    def get_bode_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot this DataSet as a Bode plot: the frequencies, the absolute magnitudes of the impedances, and the negative phase angles/shifts of the impedances in degrees.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            abs(self.impedance),
            -angle(self.impedance, deg=True),
        )

    def get_drt_data(self, imaginary: bool = False) -> Tuple[ndarray, ndarray]:
        """
        Get the data necessary to plot this DRTResult as a DRT plot: the time constants and the corresponding gamma values.

        Parameters
        ----------
        imaginary: bool = False
            Get the imaginary gamma (non-empty only when using the BHT method).

        Returns
        -------
        Tuple[ndarray, ndarray]
        """
        gamma: ndarray = self.gamma if not imaginary else self.imaginary_gamma
        if not gamma.any():
            return (
                array([]),
                array([]),
            )
        return (
            self.tau,
            gamma,
        )

    def get_drt_credible_intervals(self) -> Tuple[ndarray, ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the Bayesian credible intervals for this DRTResult: the time constants, the mean gamma values, the lower bound gamma values, and the upper bound gamma values.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray, ndarray]
        """
        if not self.mean_gamma.any():
            return (
                array([]),
                array([]),
                array([]),
                array([]),
            )
        return (
            self.tau,
            self.mean_gamma,
            self.lower_bound,
            self.upper_bound,
        )

    def get_residual_data(self) -> Tuple[ndarray, ndarray, ndarray]:
        """
        Get the data necessary to plot the relative residuals for this DRTResult: the frequencies, the relative residuals for the real parts of the impedances in percents, and the relative residuals for the imaginary parts of the impedances in percents.

        Returns
        -------
        Tuple[ndarray, ndarray, ndarray]
        """
        return (
            self.frequency,
            self.real_residual * 100,
            self.imaginary_residual * 100,
        )

    def get_scores(self) -> Dict[str, complex]:
        """
        Get the scores (BHT method) for the data set.
        The scores are represented as complex values where the real and imaginary parts have magnitudes ranging from 0.0 to 1.0.
        A consistent impedance spectrum should score high.

        Returns
        -------
        Dict[str, complex]
        """
        return self.scores

    def get_score_dataframe(self, latex_labels: bool = False) -> Optional[DataFrame]:
        """
        Get the scores (BHT) method for the data set as a pandas.DataFrame object that can be used to generate, e.g., a Markdown table.

        Parameters
        ----------
        latex_labels: bool = False
            Whether or not to use LaTeX macros in the labels.

        Returns
        -------
        Optional[DataFrame]
        """
        if not self.scores:
            return None
        return DataFrame.from_dict(
            {
                "Score": [
                    "Mean" if not latex_labels else r"$s_\mu$",
                    "Residuals, 1 sigma" if not latex_labels else r"$s_{1\sigma}$",
                    "Residuals, 2 sigma" if not latex_labels else r"$s_{2\sigma}$",
                    "Residuals, 3 sigma" if not latex_labels else r"$s_{3\sigma}$",
                    "Hellinger distance" if not latex_labels else r"$s_{\rm HD}$",
                    "Jensen-Shannon distance" if not latex_labels else r"$s_{\rm JSD}$",
                ],
                ("Real (%)" if not latex_labels else r"Real (\%)"): [
                    self.scores["mean"].real * 100,
                    self.scores["residuals_1sigma"].real * 100,
                    self.scores["residuals_2sigma"].real * 100,
                    self.scores["residuals_3sigma"].real * 100,
                    self.scores["hellinger_distance"].real * 100,
                    self.scores["jensen_shannon_distance"].real * 100,
                ],
                ("Imaginary (%)" if not latex_labels else r"Imaginary (\%)"): [
                    self.scores["mean"].imag * 100,
                    self.scores["residuals_1sigma"].imag * 100,
                    self.scores["residuals_2sigma"].imag * 100,
                    self.scores["residuals_3sigma"].imag * 100,
                    self.scores["hellinger_distance"].imag * 100,
                    self.scores["jensen_shannon_distance"].imag * 100,
                ],
            }
        )
